# process all internal html: \tooltip, textbf, \href, \shinyInternalLink, etc.
# called from htmlify()
# tex must be of length one
processText = function(tex, texRange, globalInfoEnv) {
  assure(length(tex)==1, 221956)

  # Separate mathJax to avoid, e.g., "\bf" ==> "<strong>" inside math (and change "$foo$" to "\(foo\)")
  tex = gsub("[\\][$]", "\v", tex)  # single "\v" character assumed not in text, and used as a placeholder
  temp = gregexpr("([$])?([$])([^$])+([$])([$])?", tex)[[1]]
  hasMathJax = length(temp)>1 || temp != -1
  if (hasMathJax) {
    vec = NULL
    vBegin = as.numeric(temp)
    vEnd = vBegin + attr(temp, "match.length") - 1
    vLen = length(vBegin)
    texLen = nchar(tex)
    if (vBegin[1] > 1) vec = substring(tex, 1, vBegin[1]-1)
    for (i in 1:vLen) {
      tempS = substring(tex, vBegin[i], vEnd[i])
      tempL = nchar(tempS)
      if (substring(tempS,2,2) != "$" && substring(tempS, tempL-1, tempL-1) != "$")
        tempS = paste0("\\(", substring(tempS, 2, tempL-1), "\\)")
      vec = c(vec, tempS)
      if (i < vLen) {
        if (vEnd[i+1] > vBegin[i]+1) {
          vec = c(vec, substring(tex, vEnd[i]+1, vBegin[i+1]-1))
        }
      } else { # at last mathJax string
        if (vEnd[i] < texLen) {
          vec = c(vec, substring(tex, vEnd[i]+1, texLen))
        }
      }
    } # end for each mathJax string
    tex = vec
  } # end if any mathJax

  # hold out math from the next set of changes
  if (hasMathJax) {
    holdPos = grep("^([$$]|[\\][(]).*([$$]|[\\][)])$", tex)
    holdTex = tex[holdPos]
    tex[holdPos] = "\f" # formfeed assumed not in text, and used as a placeholder
  }
  tex = paste(tex, collapse='')

  # LaTex and shinyTex commands that do not generate Shiny code  !!!! ????? DOES NOT HANDLE NESTING !!!! ????
  # handle \textbf, \textem, \shinyTip, \shinyLink   (with 1 to 2 "options" in braces)
  tex = gsub("(``|'')", '"', tex)
  tex = gsub("[\\](textbfem|textembf)[[:blank:]]*[{]([^{}]+?)[}]", "<strong><em>\\2</em></strong>", tex)
  tex = gsub("[\\]textbf[[:blank:]]*[{]([^{}]+?)[}]", "<strong>\\1</strong>", tex)
  tex = gsub("[\\]textem[[:blank:]]*[{]([^{}]+?)[}]", "<em>\\1</em>", tex)
  tex = gsub("[\\]code[[:blank:]]*[{]([^{}]+?)[}]", "<span id='shinyTexCode'>\\1</span>", tex)
  tex = gsub("[\\]textsc[[:blank:]]*[{]([^{}]+?)[}]", "<span style='font-variant:small-caps'>\\1</span>", tex)
  tex = gsub("[\\]tooltip([^[:alpha:]])", "\\\\shinyTip\\1", tex)
  tex = gsub("[\\]href([^[:alpha:]])", "\\\\shinyLink\\1", tex)
  tex = gsub("[\\]h([1-4])[[:blank:]]*[{]([^{}]*)[}]", "<h\\1>\\2</h\\1>", tex)
  tex = gsub("[\\]tooltipHref([^[:alpha:]])", "\\\\shinyLink\\1", tex)
  tex = gsub("[\\]shinyTip[[:blank:]]*[{]([^{}]+?)[}][[:blank:]]*[{]([^{}]+?)[}]",
             "<span title='\\2'><strong><em>\\1</em></strong></span>", tex)
  tex = gsub("[\\]shinyLink[[:blank:]]*[{]([^{}]+?)[}][[:blank:]]*[{]([^[:blank:]{}]+?)[}][[:blank:]]*[{]([^{}]+?)[}]",
             "<span title='\\2'><a href='\\3' target='_self'>\\1</a></span>", tex)
  tex = gsub("[\\]shinyLink[[:blank:]]*[{]([^{}]+?)[}][[:blank:]]*[{]([^{}]+?)[}]",
             "<a href='\\2' target='_self'>\\1</a>", tex)
  tex = gsub("[\\]shinyLinkNew[[:blank:]]*[{]([^{}]+?)[}][[:blank:]]*[{]([^[:blank:]{}]+?)[}][[:blank:]]*[{]([^{}]+?)[}]",
             "<span title='\\2'><a href='\\3' target='_blank'>\\1</a></span>", tex)
  tex = gsub("[\\]shinyLinkNew[[:blank:]]*[{]([^{}]+?)[}][[:blank:]{}]*[{]([^[{}]+?)[}]",
             "<a href='\\2' target='_blank'>\\1</a>", tex)
  tex = gsub("[\\]shinyMailto[[:blank:]]*[{]([^{}]+?)[}][[:blank:]]*[{]([^[:blank:]{}]+?)[}][[:blank:]]*[{]([^{}]+?)[}]",
             "<span><a href='mailto:\\2?subject=\\3'>\\1</a></span>", tex)

  # ??inside tooltips, multiple spaces show as multiple spaces
  tex = gsub("[[:blank:]]{2,}", " ", tex)
  tex = gsub("[{][[:blank:]]*[\\](bfem|embf)[[:blank:]]+(.*?)[}]", "<strong><em>\\1</em></strong>", tex)
  tex = gsub("[{][[:blank:]]*[\\]bf[[:blank:]]+(.*?)[}]", "<strong>\\1</strong>", tex)
  tex = gsub("[{][[:blank:]]*[\\]em[[:blank:]]+(.*?)[}]", "<em>\\1</em>", tex)
  tex = gsub("[{][[:blank:]]*[\\]sc[[:blank:]]+(.*?)[}]", "<span style='font-variant:small-caps'>\\1</span>", tex)
  tex = gsub("[\\]item[[:blank:]]+", "<li> ", tex)

  #lineNumOffset = gsub("([@])([0-9]+)(.*)", "\\2", texRange)
  texRangeX = substring(texRange, 2) # remove leading "@"

  # spaces{#} for horizontal spaces
  locs = gregexpr("[\\]spaces[[:space:]]*[{].+?[}]", tex)[[1]]
  if (locs[1] != -1) {
    lens = attr(locs, "match.length")
    for (item in length(locs):1) {
      P1 = locs[item]
      P2 = locs[item]+lens[item]-1
      cmd = substring(tex, P1, P2)
      num = suppressWarnings(as.numeric(gsub(".*[{](.+?)[}]", "\\1", cmd)))
      if (is.na(num)) stop("bad number for \\spaces{} near line ", texRangeX)
      str = paste0("<span>", paste(rep("&nbsp;", num), collapse=""), "</span>")
      tex = paste0(substring(tex, 1, P1-1), str, substring(tex, P2+1))
    }
  }
  
  # shinyInternalLoc and shinyInternalLink
  locs = gregexpr("[\\]shinyInternal(Link|Loc)[[:space:]]*[{].+?[}]", tex)[[1]]
  internalLinks = getGlobal("internalLinks", globalInfoEnv)
  internalLocs = getGlobal("internalLocs", globalInfoEnv)
  if (locs[1] != -1) {
    lens = attr(locs, "match.length")
    for (item in length(locs):1) {
      P1 = locs[item]
      P2 = locs[item]+lens[item]-1
      cmd = substring(tex, P1, P2)
      prs = parseCommand(cmd, texRangeX)
      type = prs$type
      options = prs$options
      if (type=="shinyInternalLink") {
        options = parseOptions(type, options, 
                               required = "name", requiredTypes = "c",
                               optional = "text", optionalTypes = "c",
                               lineNumOffset=texRangeX)
        if (is.null(options$text)) options$text = options$name
        str = paste0('<a href=\'#', make.names(options$name), '\' id=\'shinyTexInternalLink\'>', options$text, '</a>')
        if (is.null(internalLinks)) {
          internalLinks = data.frame(name=I(options$name), lineNum=texRangeX)
        } else {
          internalLinks = rbind(internalLinks, data.frame(name=I(options$name), lineNum=texRangeX))
        }
        assign("internalLinks", internalLinks, envir=globalInfoEnv)
      } else {  # shinyInternalLoc
        options = parseOptions(type, options, 
                               required = "name", requiredTypes = "c",
                               optional = NULL, optionalTypes = NULL,
                               lineNumOffset=texRangeX)
        str = paste0('<a name=', make.names(options$name), '></a>')
        if (!is.null(internalLocs) && options$name %in% internalLocs)
          stop("'\\internalLoc{name=", options$name, "} " , texRange, " duplicates earlier usage")
        assign("internalLocs", c(internalLocs, options$name), envir=globalInfoEnv)
      }
      tex = paste0(substring(tex, 1, P1-1), str, substring(tex, P2+1))
    } # end for each InternalLoc/Link
  } # end if any InternalLoc/Links
  
  # final check for Latex commands
  finalCheck = gregexpr("[\\]([[:alpha:]]{2,})", tex)[[1]]
  #finalOK 
  if (finalCheck[1] != -1) {
    strings = substring(tex, finalCheck, finalCheck+attr(finalCheck,"match.length")-1)
    #worries = grep("[\\]shiny", strings, invert=TRUE)
    #if (any(worries)) {
      worries = 1:length(strings)  #????!!! remove ????!!!!!
      warning(texRange, ' un-handled LaTeX commands: ', strings[worries][1], 
              ifelse(length(worries)==1, '', paste(' (and', length(worries)-1, 'others)')))
    #}
  } # end checking for un-handle LaTeX
  
  # restore mathJax
  if (hasMathJax) {
    formfeed = gregexpr("\f", tex)[[1]]
    assure(length(formfeed) == length(holdPos), 83645)
    for (i in 1:length(holdPos)) tex = sub("\f", holdTex[i], tex, fixed=TRUE)
  }
  
  
  return(tex)
} # end processText()

