# rFeedback() handles the \shinyRFeedback environment for innerCode()


rFeedback = function(text, inputId, options, lineNumOffset, globalInfoEnv) {
  if (inputId=="" || make.names(inputId)!=inputId) {
    stop("@ ", lineNumOffset, 
         " first argument of \\begin{shinyRFeedback} must be a valid R name (inputId)")
  }
  
  options = parseCommand(paste0("shinyRFeedback(second_curly_brace_set){", options, "}"), 
                         lineNumOffset)$options

  options = parseOptions("shinyRFeedback", options, 
                         required=NULL, requiredTypes=NULL,
                         optional=c("nrows","ncols","label","feedbackLabel",
                                    "submitLabel","allowSemicolons",
                                    "vars", "carryOverVars",
                                    "monitorAll", "monitorErrors"), 
                         optionalTypes=c("n","n","c","c","c","l","c","l","l","l"),
                         lineNumOffset=lineNumOffset,
                         defaults=list(nrows=1, ncols=30, 
                                       label=ifelse(is.null(options$nrows) || options$nrows==1,
                                                    "R code prompt (>):",
                                                    "R code input:"),
                                       feedbackLabel="Feedback:", 
                                       submitLabel="Submit R Code",
                                       allowSemicolons=FALSE,
                                       vars="",
                                       carryOverVars=FALSE,
                                       monitorAll=FALSE,
                                       monitorErrors=TRUE))
  if (options$nrows > 1) options$allowSemicolons = TRUE
  
  lineCnt = options$nrows
  if (lineCnt<1 || lineCnt>30 || lineCnt!=round(lineCnt)) 
    stop("@ ", lineNumOffset, " 'lines' option of \\shinyRFeedback must be from 1 to 30")
  
  # Update global info
  priorRF = getGlobal("RFeedback", globalInfoEnv)
  if (is.null(priorRF)) {
    assign("RFeedback", data.frame(id=inputId, lineNum=lineNumOffset, 
                                   allowSemicolons=options$allowSemicolons, vars=options$vars), 
           envir=globalInfoEnv)
                                   
  } else {
    Sel = inputId==priorRF$id
    if (any(Sel)) {
      stop("\\shinyRFeedback id '",  inputId, "' @ both ", priorRF$lineNum[Sel], " and ", lineNumOffset)
    }
    assign("RFeedback", rbind(priorRF, data.frame(id=inputId, lineNum=lineNumOffset,
                                                  allowSemicolons=options$allowSemicolons, vars=options$vars)), 
           envir=globalInfoEnv)
  }
  

  ####################### Set User Interface code for this RFeedback block ###########################
  ########## To get textEnter to respond only at Enter, go to
  ########## C:\Users\hseltman\Documents\R\win-library\3.2\shiny\www\shared
  ########## and erase "$(a).on("keyup.textInputBinding input.textInputBinding",function(a){b(!0)}),"
  if (lineCnt == 1) {
    feedInput = paste0('textInput(inputId="readRCode', inputId, '", label="',
                       options$label, '", width="', options$ncols, 'em"),')
  } else {
    feedInput = c(paste0('HTML("<p><strong>', options$label, '</strong></p>"),'),
                  paste0('textareaInput(inputId="readRCode', inputId, '",',
                         'nrows=', lineCnt, ', ncols=', options$ncols, '),'),
                  'HTML(\'<div style="padding:1ex"></div>\'),',
                  #paste0('HTML("<p><strong>', options$feedbackLabel, '</strong></p>"),'),
                  paste0('actionButton(inputId="readRSubmit', inputId,
                         '", label="', options$submitLabel,'"),'))
  }
  uiCode = c(feedInput,
             'HTML(\'<div style="padding:1ex"></div>\'),',
             paste0('HTML("<p><strong>', options$feedbackLabel, '</strong></p>"),'),
             paste0('verbatimTextOutput(outputId="rCodeFeedbackOut', inputId, '"),'),
             'HTML(\'<div style="padding:1ex"></div>\'),')

    
  ## Set server code ##
  # testing of 'env' argument at compile time
  if (options$vars != "") {
    tempVars = try(eval(parse(text=paste0('list(', options$vars, ')'))))
    if (is(tempVars, "try-error")) stop("bad 'env=' for feedback '", inputId, " @ ", lineNumOffset)
    if (is.null(names(tempVars)) || any(names(tempVars)==""))
      stop("all elements must be named in 'env=' for feedback '", inputId, " @ ", lineNumOffset)
  }
  if (options$carryOverVars) {
    serverCode = c(paste0('feedbackEnv', inputId, ' = list2env(list(', options$vars, 
                          '), parent=as.environment("package:stats"))'))
  } else serverCode = NULL
  serverCode = c(serverCode, 
                 paste0('readRCodeRtn', inputId, ' = reactive({'))
  
  if (options$carryOverVars == FALSE) {
    serverCode = c(serverCode, 
                   paste0('  feedbackEnv', inputId, ' = list2env(list(', options$vars, 
                          '), parent=as.environment("package:stats"))'))
  }
  
  serverCode = c(serverCode,
                 paste0('  rtn = runR(input$readRCode', inputId, ', feedbackEnv', inputId,
                                      ', allowSemicolons=', as.character(options$allowSemicolons), ')'),
                 paste0('  vars = eapply(feedbackEnv', inputId, ', identity)'),
                 '  return(c(rtn, list(VARS=vars)))',
                 '})')
  
  
  serverCode = c(serverCode,
                 feedbackCommon(inputId, text, isHidden=FALSE, options, lineNumOffset, globalInfoEnv))
    
  return(list(uiCode=uiCode, serverCode=serverCode))
}
