# start the "do list"
# each element is a list of four elements: 
# 1) a string containing UI code to be written to ui.R or 
#    a c(start,end) of line numbers for deeper processing
# 2) a string containing server code to be written to server.R or 
#    NULL if none
# 3) the list position of the next element (cannot be called keyword 'next')
# 4) the indent level
#
startDoList = function(tex, defs, tabTitles, mainStartEnd, tabbing) {
  # Construct title
  appTitle = paste0("HTML(\"<span style='font-weight:", defs$titleFontWeight,
                    ";color:", defs$titleColor,
                    ";font-size:", defs$titleFontPercent, '%',
                    ";font-family:", defs$titleFont,
                    "'>", defs$title, "</span>\")")

  appCSS = c("  singleton(tags$link(rel = 'stylesheet', type = 'text/css', href = 'shinyTex.css')),",
             paste0("  singleton(tags$style(type = 'text/css', 'body {background-color: ", defs$backgroundColor, ";}')),"),
             paste0("  singleton(tags$style(type = 'text/css', 'pre {background-color: ", defs$verbatimColor, ";}')),"),
             paste0("  singleton(tags$style(type = 'text/css', '#shinyTexInternalLink{background-color: ", 
                    defs$internalLinkBackgroundColor, "; color: ", defs$internalLinkColor, "; font: ",
                    defs$internalLinkFont, "}')),"),
             paste0("  singleton(tags$style(type = 'text/css', '#shinyTexCode{background-color: ", 
                    defs$codeBackgroundColor, "; color: ", defs$codeColor, "; font: ",
                    defs$codeFont, "}')),"))
  if (!is.null(defs$headerBackground)) {
    appCSS = c(appCSS,
               paste0("  singleton(tags$style(type = 'text/css', '.navbar {background-color: ", defs$headerBackground, ";}')),"))
  }
  if (!is.null(defs$titleBackground)) {
    appCSS = c(appCSS,
               paste0("  singleton(tags$style(type = 'text/css', '.navbar-brand {background-color: ", defs$titleBackground, ";}')),"))
  }
  if (!is.null(defs$navbarBackground)) {
    appCSS = c(appCSS,
               paste0("  singleton(tags$style(type = 'text/css', '.navbar-nav {background-color: ", defs$navbarBackground, ";}')),"))
  }
  
  # Construct list
  if (!tabbing && nrow(mainStartEnd) == 1) { # unstructured single page
    lst = vector("list", 3)
    lst[[1]] = list(uiCode = c(paste0('# Shiny ui.R built by shinyTex ', Sys.time(), 
                                      ifelse(is.null(defs$author), '', paste(' by ', defs$author)),
                                      ifelse(is.null(defs$title), '', paste('\n# ', defs$title))),
                               '', 'library(shiny)', 'library(stringr)',
                               ifelse(is.null(defs$theme), '', 'library(shinythemes)'),
                               'source("shinyTexAux.R")', '',
                               '# begin UI definition',
                               paste0('fluidPage(', ifelse(is.null(defs$theme), '', 
                                                           paste0('theme = shinytheme("', defs$theme, '"),'))),
                               paste0('  titlePanel(', appTitle, ', "', defs$title, '"),'),
                               #'  fluidPanel(withMathJax(),',
                               '  withMathJax(),',
                               appCSS),
                    serverCode=NA, nextElement = 2, indent = 0)
    lst[[2]] = list(uiCode = c(mainStartEnd[1,]),
                    serverCode=NA, nextElement=3, indent = 1)
    lst[[3]] = list(uiCode = ') # end fluidPage()',
                    serverCode=NA, nextElement = NA, indent = 0)
  } else if (!tabbing && nrow(mainStartEnd) > 1) { # single page with sidebar and main
    lst = vector("list", 5)
    lst[[1]] = list(uiCode = c(paste('# Shiny ui.R built by shinyTex', Sys.time(), 
                                     ifelse(is.null(defs$author), '', paste('by', defs$author)),
                                     ifelse(is.null(defs$title), '', paste('\n#', defs$title))),
                               '', 'library(shiny)', 'library(stringr)', 
                               ifelse(is.null(defs$theme), '', 'library(shinythemes)'),
                               'source("shinyTexAux.R")', '',
                               '# begin UI definition',
                               paste0('fluidPage(', ifelse(is.null(defs$theme), '', 
                                                           paste0('theme = shinytheme("', defs$theme, '"),'))),
                               paste0('  titlePanel(', appTitle, ', "', defs$title, '"),'),
                               '  sidebarLayout(', '  sidebarPanel('),
                    serverCode=NA, nextElement = 2, indent = 0)
    lst[[2]] = list(uiCode = mainStartEnd[1,], 
                    serverCode=NA, nextElement=3, indent = 1)
    lst[[3]] = list(uiCode = c("),", "", "mainPanel(", 
                               "  withMathJax(),",
                               appCSS),
                    serverCode=NA, nextElement = 4, indent = 0)
    lst[[4]] = list(uiCode = mainStartEnd[2,], 
                    serverCode=NA, nextElement=5, indent = 1)
    lst[[5]] = list(uiCode = '))) # end mainPanel() and fluidPage()',
                    serverCode=NA, nextElement = NA, indent = 0)
  } else { # if tabbing
    nTabs = nrow(mainStartEnd)
    lst = vector("list", 3*nTabs+2)

    # setup each do list element for the tabs
    lst[[1]] = list(uiCode = c(paste('# shiny.ui built by shinyTex', Sys.time(), 
                                     ifelse(is.null(defs$author), '', paste('by', defs$author))),
                               '', 'library(shiny)', 'library(stringr)',
                               ifelse(is.null(defs$theme), '', 'library(shinythemes)'),
                               'source("shinyTexAux.R")', '',
                               '# begin UI definition',
                               paste0('navbarPage(', ifelse(is.null(defs$theme), '', 
                                                           paste0('theme = shinytheme("', defs$theme, '"),'))),
                               paste0('  title =', appTitle, ',')),
                    serverCode=NA, nextElement = 2, indent = 0)
    for (tab in 1:nTabs) {
      offset = 3*(tab-1) + 1
      code =  c('tabPanel(', 
                paste0('  title = "', tabTitles[tab], '",'))
      if (tab == 1) code = c(code, 
                             "  withMathJax(),",
                             "  singleton(tags$head(tags$script(src = 'tabJumpCallback.js'))),",
                             appCSS)
      lst[[offset + 1]] = list(uiCode = code, serverCode=NA, nextElement = offset+2, indent = 1)
      lst[[offset + 2]] = list(uiCode = c(mainStartEnd[tab,1]+1, mainStartEnd[tab,2]-1), 
                               serverCode=NA, nextElement = offset+3, indent = 2)
      if (tab < nTabs) {
        lst[[offset + 3]] = list(uiCode = c(paste('), # end tab for', tabTitles[tab]), ''),
                                 serverCode=NA, nextElement = offset+4, indent = 1)
      } else {
        lst[[offset + 3]] = list(uiCode = c(paste('  ), # end tab for', tabTitles[tab]), '',
                                            paste0('  windowTitle="', defs$title, '"')),
                                 serverCode=NA, nextElement = offset+4, indent = 0)
      }
    } # end for each tab
    lst[[3*nTabs+2]] = list(uiCode = ') # end navbarPage',  serverCode=NA, nextElement = NA, indent = 0)

    # construct tabJumpSetup.js
    if (!file.exists("www/tabJumpCallback.js")) {
      write(c('Shiny.addCustomMessageHandler("tabJumpCallback",',
              '  function(tabName) {',
              '    $("a[data-value=" + "\'" + tabName + "\']").click();',
              '    window.scrollTo(0,0);',
              '  });'), 
            file = "www/tabJumpCallback.js")
    }
  } # end tabbing

  # Allow text areas (construct textarea.js)
  if (!file.exists("www/textarea.js")) {
    write(c('$(document).on("click", "textarea.inputTextarea", function(evt) {',
            '  var el = $(evt.target);',
            '  el.trigger("change");',
            '});',
            'var textareaInputBinding = new Shiny.InputBinding();',
            '$.extend(textareaInputBinding, {',
            '  find: function(scope) {',
            '    return $(scope).find(".textareaInput");',
            '  },',
            '  getValue: function(el) {',
            '     return $(el).text();',
            '  },',
            '  setValue: function(el, value) {',
            '    $(el).text(value);',
            '  },',
            '  subscribe: function(el, callback) {',
            '    $(el).on("change.textareaInputBinding", function(e) {',
            '      callback();',
            '    });',
            '  },',
            '  unsubscribe: function(el) {',
            '    $(el).off(".textareaInputBinding");',
            '  }',
            '});',
            'Shiny.inputBindings.register(textareaInputBinding);'),
          file = "www/textarea.js")
    }

    
  # Allow text input with explict evaluation time (construct inputTextEnter.js)
  if (!file.exists("www/textInputEnter.js")) {
    write(c('var textInputEnterBinding = new Shiny.InputBinding();',
            '$.extend(textInputEnterBinding, {',
            '  find: function(scope) {',
            '    return $(scope).find(".textInputEnter");',
            '  },',
            '  getId: function(el) {',
            '    return $(el).attr("id")',
            '  },',
            '  getValue: function(el) {',
            '    return el.value;',
            '  },',
            '  setValue: function(el, value) {',
            '    el.value = value;',
            '  },',
            '  subscribe: function(el, callback) {',
            '    $(el).on("keyup.textInputBinding input.textInputBinding", function(event) {',
            '      if(event.keyCode == 13) {',
            '        callback()',
            '      }',
            '    });',
            '    $(el).on("focusout.textInputEnterBinding", function(event) {', 
            '      callback();',
            '    });',
            '  },',
            '  unsubscribe: function(el) {',
            '    $(el).off(".textInputEnterBinding");',
            '  },',
            '  receiveMessage: function(el, data) {',
            '    if (data.hasOwnProperty("value"))',
            '      this.setValue(el, data.value);',
            '    if (data.hasOwnProperty("label"))',
            '      $(el).parent().find("label[for=" + el.id + "]").text(data.label);',
            '    $(el).trigger("change");',
            '  },',
            '  getState: function(el) {',
            '    return {',
            '      label: $(el).parent().find("label[for=" + el.id + "]").text(),',
            '      value: el.value',
            '    };',
            '  },',
            '  getRatePolicy: function() {',
            '    return {',
            '      policy: "debounce",',
            '      delay: 250',
            '    };',
            '  }',
            '});',
            'Shiny.inputBindings.register(textInputEnterBinding);'),
          file = "www/textInputEnter.js")
    }
    


  
  # Write start/end server code
  lst[[1]]$serverCode = c(paste('# Shiny server.R built by shinyTex', Sys.time(), 
                                ifelse(is.null(defs$author), '', paste('by', defs$author)),
                                ifelse(is.null(defs$title), '', paste('\n#', defs$title))),
                          '', 'library(shiny)', 'library(stringr)',
                          #'source("shinyTexAux.R")', 
                          '',
                          'shinyServer(function(input, output, session) {')

  lst[[length(lst)]]$serverCode = '}) # end shinyServer()'
  
  return(lst)
  
} # end startDoList
